/*
 * puzzle.c
 *
 *
 * reconstruction of tree topologies with quartet puzzling
 *
 * by Korbinian Strimmer and Arndt von Haeseler (1995)
 *
 */


#define MAIN_MODULE 1

#include "puzzle.h"


/******************* internal functions ****************************************/

/* set options interactively */
void setoptions()
{	
	char inputline[10];	


	/* defaults */
	Info_optn   = FALSE; /* show status of computation       */
	pmode_optn  = 1;     /* use all good quartets            */
	Approx_optn = TRUE;  /* approximate likelihoods          */
	Numtrial = 100;      /* number of quartet puzzling trees */


#ifdef NUC	
	AlphaYR = 1.0;
	AlphaBeta = 4.0;
	Frequ_optn = TRUE;
#endif

#ifdef PROT
	Dayhf_optn = FALSE;
	Jtt_optn = TRUE;
	Poisn_optn = FALSE;
	Frequ_optn = FALSE;
#endif	

	fprintf(stderr, "\nReconstructing tree topologies with quartet puzzling\n");

 	for (;;){
			
		fprintf(stderr, "\n\nSettings for this run of Puzzle version %s (%s):\n\n",
			VERSION, MODEL);	
		fprintf(stderr, "p           Quartet puzzling algorithm?  ");
		if (pmode_optn == 0) fprintf(stderr, "Use all quartets\n");
		if (pmode_optn == 1) fprintf(stderr, "Use all good quartets\n");
#ifdef NUC			
		fprintf(stderr, "m          Model of sequence evolution?  ");
		if(AlphaBeta == 1.0 && AlphaYR == 1.0) fprintf(stderr, "%s\n",
			(Frequ_optn ? "F (Felsenstein 1981)" : "JC (Jukes-Cantor 1969)"));
		if((AlphaBeta != 1.0 && AlphaYR == 1.0)) fprintf(stderr, "%s\n",
			(Frequ_optn ? "HKY (Hasegawa et al. 1985)" : "K2P (Kimura 1980)"));	
		if(AlphaYR != 1.0) fprintf(stderr, "%s\n", "TN (Tamura-Nei 1993)");
		if (AlphaYR != 1.0 || AlphaBeta != 1.0)
			fprintf(stderr, "t        Transition/transversion ratio?  %.2f\n",
				AlphaBeta / 2);	
		if (AlphaYR != 1.0)
			fprintf(stderr, "r      Y-transition/R-transition ratio?  %.2f\n",
				AlphaYR);
#endif

#ifdef PROT
		fprintf(stderr, "m          Model of sequence evolution?  ");
		if (Dayhf_optn) fprintf(stderr, "Dayhoff (Dayhoff et al. 1978)\n");	
		if (Jtt_optn) fprintf(stderr, "JTT (Jones et al. 1992)\n");
		if (Poisn_optn) fprintf(stderr, "%s\n",
			(Frequ_optn ? "F (Felsenstein 1981)" : "JC (Jukes-Cantor 1969)"));
#endif

		fprintf(stderr, "f       Use empirical base frequencies?  %s\n",
			(Frequ_optn ? "Yes" : "No") );	
		fprintf(stderr, "a       Approximate maximum likelihood?  %s\n",
				(Approx_optn ? "Yes" : "No"));
		fprintf(stderr, "n     Number of quartet puzzling trees?  %d\n",
				Numtrial);
		fprintf(stderr, "i          Show details of computation?  %s\n",
				(Info_optn ? "Yes" : "No"));
		
		fprintf(stderr, "\nStart with analysis? ");
		fprintf(stderr, "(type y or choose from the menu) ");
		
		gets(inputline);

		if(tolower(inputline[0]) == 'y')
		break;

		switch (tolower(inputline[0])) {

			case 'a':	Approx_optn = 1 - Approx_optn;
						break;
			
			case 'f':	Frequ_optn = 1 - Frequ_optn;
						break;
			
			case 'i':	Info_optn = 1 - Info_optn;
						break;

			case 'n':	fprintf(stderr, "\n\nNumber of quartet puzzling trees: ");
						scanf("%d", &Numtrial);
						gets(inputline);
						if (Numtrial < 1) {
							fprintf(stderr,
								"\n\nAt least one tree has to be generated!\n");
							Numtrial = 100;
						}
						break;

			case 'p':	pmode_optn++;
						if (pmode_optn > 1) pmode_optn = 0;
						break;

#ifdef NUC			
			case 'm':	if(AlphaBeta == 1.0 && AlphaYR == 1.0) {/* -> K2P/HKY */
							AlphaBeta = 4.0;
							AlphaYR = 1.0;
							break;
						}
						if(AlphaBeta != 1.0 && AlphaYR == 1.0) {/* -> TN */
							AlphaBeta = 4.0;
							AlphaYR = 2.0;
							break;
						}
						if(AlphaYR != 1.0) { /* -> JC */
							AlphaBeta = 1.0;
							AlphaYR = 1.0;
							break;
						}
						break;
						
			case 't':	if (AlphaYR == 1.0 && AlphaBeta == 1.0) {
							fprintf(stderr,"\n\nThis is not a possible option!\n");
							break;
						} else {
							fprintf(stderr, "\n\nExpected transition/transversion ratio: ");
							scanf("%lf", &AlphaBeta);
							AlphaBeta = 2 * AlphaBeta;
							gets(inputline);
							if (AlphaBeta < 0) {
								fprintf(stderr, "\n\nNo negative values for the transition/transversion ratio allowed!\n");
								AlphaBeta = 4.0;
							}
							if (AlphaBeta == 0) {
								fprintf(stderr, "\n\nNo zero value for the transition/transversion ratio allowed!\n");
								AlphaBeta = 4.0;
							}
							break;
						}
						break;
		
			case 'r':	if (AlphaYR == 1.0) {
							fprintf(stderr,"\n\nThis is not a possible option!\n");
							break;
						} else {
							fprintf(stderr, "\n\nPyrimidine transition/Purin transition ratio: ");
							scanf("%lf", &AlphaYR);
							gets(inputline);
							if (AlphaYR < 0) {
								fprintf(stderr, "\n\nNo negative values for the Y-transition/R-transition ratio allowed!\n");
								AlphaYR = 2.0;
							}
							if (AlphaYR == 0) {
								fprintf(stderr, "\n\nNo zero value for the Y-transition/R-transition ratio allowed!\n");
								AlphaYR = 2.0;
							}
							break;
						}
						break;
#endif

#ifdef PROT
			case 'm':	if (Dayhf_optn) { /* -> JC/F */
							Dayhf_optn = FALSE;
							Jtt_optn = FALSE;
							Poisn_optn = TRUE;
							break;
						}
						if (Poisn_optn) { /* -> JTT model */
							Dayhf_optn = FALSE;
							Jtt_optn = TRUE;
							Poisn_optn = FALSE;
							break;
						}
						if (Jtt_optn) { /* -> Dayhoff model */
							Dayhf_optn = TRUE;
							Jtt_optn = FALSE;
							Poisn_optn = FALSE;
							break;
						}
						break;

#endif			
		default:	fprintf(stderr, "\n\nThis is not a possible option!\n");
		}
	}
	fprintf(stderr, "\n\n");
}


/* open all needed files */
void opentreefile()
{	
	char inputline[64];

	fprintf(stderr, "\n\nPlease enter the name for the output file: ");
	gets(inputline);
	if ((treefp = fopen(inputline, "w")) == NULL) {
		fprintf(stderr, "\nERROR: I can't create the output file: %s\n", inputline);
		fprintf(stderr, "Please enter another name for this file: ");
		gets(inputline);
		if ((treefp = fopen(inputline, "w")) == NULL) {
			fprintf(stderr, "\nERROR: I can't create this file either!\n");
			exit(1);
		}
	}
	fprintf(stderr, "\n");
}


/* close needed files */
void closetreefile()
{	
	fclose(treefp);
}



/******************************** main module *********************************/


main()
{	
	int i, j, a, b, c, k, ID, ii, jj, badquartet;
	unsigned long nqused, nqanalysed, trial;
	time_t starttime, stoptime;
	double diff12, diff13, diff23;
#ifdef NUC
	double AlphaYBeta, AlphaRBeta, PiA, PiC, PiG, PiT, PiR, PiY;
#endif


	/* set options in interaction with the user */
	setoptions();

	/* initialize random numbers generator */
	initrandom();
	
	/* initialise ML */
	fprintf(stderr, "\nINFO: Reading and processing sequence input data\n");
	qmlstart();

#ifdef NUC	
	/* compute transition-transversion ratio */
	PiA = Freqtpm[2];
	PiC = Freqtpm[1];
	PiG = Freqtpm[3];
	PiT = Freqtpm[0];
	PiR = PiA + PiG;
	PiY = PiC + PiT;
	AlphaRBeta = 2 * AlphaBeta / (1 + AlphaYR);
	AlphaYBeta = AlphaRBeta * AlphaYR;	
	fprintf(stderr, "\nINFO: Current expected transition/transversion ratio is %.2f\n",
		(AlphaRBeta * PiA * PiG + AlphaYBeta * PiC * PiT) / (PiR * PiY) );
#endif

	/* open treefile */
	opentreefile();

	/* variables used for randomizing input order */
	trueID = new_ivector(Maxspc);
	freeID = new_ivector(Maxspc);

	/* allocate memory for quartets */
	fprintf(stderr, "\nINFO: Preparing for quartet information storage\n");
	mallocquartets(Maxspc);

	/* start timer */
	time(&starttime);
		
	/* compute quartets */
	fprintf(stderr, "\nINFO: Computing quartet maximum likelihood values\n");
	computeallquartets();
	
	fprintf(stderr, "\nINFO: Starting quartet puzzling analysis\n");
	
	for (Currtrial = 0; Currtrial < Numtrial; Currtrial++) {
				
		/* randomize input order */
				
		for (k = 0; k < Maxspc; k++) freeID[k] = TRUE;
			if (pmode_optn == 1) { /* special care for first quartet */
				trueID[0] = startA;
				freeID[startA] = FALSE;
				trueID[1] = startB;
				freeID[startB] = FALSE;
				trueID[2] = startC;
				freeID[startC] = FALSE;
				trueID[3] = startD;
				freeID[startD] = FALSE;
				for (k = 4; k < Maxspc; k++) {
					/* random integer in the range [0, Maxspc-1-k] */
					ID = randominteger(Maxspc - k);
					i = -1;
					j = -1;
					do {
						i++;
						if (freeID[i] == TRUE) j++; 
					} while ( j != ID);
					trueID[k] = i;
					freeID[i] = FALSE;
				}
			} else {
				for (k = 0; k < Maxspc; k++) {
					/* random integer in the range [0, Maxspc-1-k] */
					ID = randominteger(Maxspc - k);
						i = -1;
						j = -1;
						do {
							i++;
							if (freeID[i] == TRUE) j++; 
						} while ( j != ID);
						trueID[k] = i;
						freeID[i] = FALSE;
				}
			}
					

			/* initialize tree */
			inittree();
			if (Info_optn) {
				fprintf(stderr, "\n  Quartet puzzling tree # %d:\n",
				Currtrial + 1);
				fprintf(stderr, "   leaf 0  > ");
				fputid10(stderr, trueID[0]);
				fprintf(stderr, " <  added  [0/0 quartets used/analysed]\n");
				fprintf(stderr, "   leaf 1  > ");
				fputid10(stderr, trueID[1]);
				fprintf(stderr, " <  added  [0/0 quartets used/analysed]\n");
				fprintf(stderr, "   leaf 2  > ");
				fputid10(stderr, trueID[2]);
				fprintf(stderr, " <  added  [0/0 quartets used/analysed]\n");
			}

			/* adding all other leafs */
			for (i = 3; i < Maxspc; i++) {
			
				/* determine place for next leaf */ 
						
				/* clear all edgeinfos */
				resetedgeinfo();
				
				/* clear counter of quartets */
				nqanalysed = 0;
				nqused = 0;
				
				
				/*
				 * core of quartet puzzling algorithm
				 */

				if (pmode_optn == 0) { /* use all quartets (original Puzzle algorithm) */

					for (a = 0; a < nextleaf - 2; a++)
						for (b = a + 1; b < nextleaf - 1; b++)
							for (c = b + 1; c < nextleaf; c++) {
						
								/*	check which two leaves out of a, b, c
									are closer related to each other than
									to leaf i. These two leaves are chooseA and chooseB */

									checkquartet(a, b, c, i);
									nqanalysed++;

									incrementedgeinfo(chooseA, chooseB, 1);
									nqused++;
							}

				}
				
				if (pmode_optn == 1) { /* use all good quartets + special care for 1st quartet */

					for (a = 0; a < nextleaf - 2; a++)
						for (b = a + 1; b < nextleaf - 1; b++)
							for (c = b + 1; c < nextleaf; c++) {
						
								/*	check which two leaves out of a, b, c
									are closer related to each other than
									to leaf i. These two leaves are chooseA and chooseB */

									badquartet = checkquartet(a, b, c, i);
									nqanalysed++;

									if (!badquartet) {
										incrementedgeinfo(chooseA, chooseB, 1);
										nqused++;
									}
							}

				}

				/* find out which edge has the lowest edgeinfo */
				minimumedgeinfo();
				
				/* add the next leaf on minedge */
				addnextleaf(minedge);
			
				if(Info_optn) {
					fprintf(stderr, "   leaf %d  > ", i);
					fputid10(stderr, trueID[i]);
					fprintf(stderr, " <  added  [%lu/%lu quartets used/analysed]\n",
						nqused, nqanalysed);
					}
				}

				/* writing treefile */
				writetree(treefp);

				/* free memory before the next trial */
				freetree();
			}

	/* close treefiles */
	closetreefile();
	fprintf(stderr, "\nINFO: All %d trees written to output file\n", Numtrial);
	
	/* stop timer */
	time(&stoptime);
	fprintf(stderr, "\nINFO: This run took %.0f seconds (= %.1f minutes)\n",
		(double) (stoptime - starttime), (stoptime - starttime)/60.);
	
	/* finish ML */
	qmlfinish();

	/* free memory */
	freequartets();
	free_ivector(trueID);
	free_ivector(freeID);

	return;
}
