/*
 * quartet.c
 *
 *
 * storing and evaluating quartet information
 *
 * by Korbinian Strimmer and Arndt von Haeseler (1995)
 *
 */


#include "puzzle.h"


/* allocate memory for quartets */
void mallocquartets(int taxa)
{	
	int a, b, c, d;
	
	/* this way of memory allocation is very efficient
	   - no quartet information is stored twice
	   - there is no unused part in quartet */
	
	/* memory allocation */
	quartet = (unsigned char ****)
		malloc((unsigned) ((taxa-3) * sizeof(unsigned char ***)));
	if (quartet == NULL) maerror("quartet [step 1]");
	for (a = 0; a < taxa - 3; a++) {
		quartet[a] = (unsigned char ***)
			malloc((unsigned) ((taxa-3-a) * sizeof(unsigned char **)));
		if (quartet[a] == NULL) maerror("quartet [step 2]");
		for (b = 0; b < taxa - 3 - a; b++) {
			quartet[a][b] = (unsigned char **)
				malloc((unsigned) ((taxa-3-a-b) * sizeof(unsigned char *)));
			if (quartet[a][b] == NULL)maerror("quartet [step 3]");
			for (c = 0; c < taxa - 3 - a - b; c++) {
				quartet[a][b][c] = (unsigned char *)
					malloc((unsigned) ((taxa-3-a-b-c) * sizeof(unsigned char)));
				if (quartet[a][b][c] == NULL) maerror("quartet [step 4]");
				for (d = 0; d < taxa - 3 - a - b - c; d++)
					quartet[a][b][c][d] = 'K';
			}
		}
	}
}


/* free quartet memory */
void freequartets()
{	
	int a, b, c;
	
	
	for (a = 0; a < Maxspc - 3; a++) 
		for (b = 0; b < Maxspc - 3 - a; b++) 
			for (c = 0; c < Maxspc - 3 - a - b; c++) 
				free(quartet[a][b][c]);
	for (a = 0; a < Maxspc - 3; a++) 
		for (b = 0; b < Maxspc - 3 - a; b++) 
				free(quartet[a][b]);
	for (a = 0; a < Maxspc - 3; a++) 
				free(quartet[a]);
	free(quartet);
}


/* checks out *all* possible quartets */
void computeallquartets()
{	
	int a, b, c, i, badquartet;
	double diff, maxdiff, diff12, diff13, diff23;
	boolean firstquartet;
	unsigned long nq, badqs;


	nq = 0;
	badqs = 0;
	firstquartet = TRUE;
	for (i = 3; i < Maxspc; i++) 
			for (a = 0; a < i - 2; a++)
					for (b = a + 1; b < i - 1; b++)
						for (c = b + 1; c < i; c++) {
							nq++;

							if(Info_optn) {
								fprintf(stderr, "\n  Checking quartet # %lu", nq);
							}
							if (Approx_optn) {
								d1 = quartet_alklhd(a,b,c,i);
								d2 = quartet_alklhd(a,c,b,i);
								d3 = quartet_alklhd(a,i,b,c);
							} else {
								d1 = quartet_lklhd(a,b,c,i);
								d2 = quartet_lklhd(a,c,b,i);
								d3 = quartet_lklhd(a,i,b,c);
							}

							if (d1 < d2) {
								if (d2 < d3) { /* d3 max */
									quartet[a][b-a-1][c-b-1][i-c-1] = 195;
									lowestml = 3;
									diff = d3 - d2;
								} else { /* d2 max */
									quartet[a][b-a-1][c-b-1][i-c-1] = 204;
									lowestml = 2;
									diff = d2 - (d1 > d3 ? d1 : d3);
								}
							} else { 
								if (d1 < d3) { /* d3 max */
									quartet[a][b-a-1][c-b-1][i-c-1] = 195;
									lowestml = 3;
									diff = d3 - d1;
								} else { /* d1 max */
									quartet[a][b-a-1][c-b-1][i-c-1] = 240;
									lowestml = 1;
									diff = d1 - (d2 > d3 ? d2 : d3);
								}
							}


							/* check whether this quartet a good quartet */

							diff12 = (d1-d2)*(d1-d2);
							diff13 = (d1-d3)*(d1-d3);
							diff23 = (d2-d3)*(d2-d3);
							badquartet = FALSE;

							if (diff12 < diff13) {
								if (diff12 < diff23) { /* diff12 min */
									/* lowestml should be 3 */
									if (lowestml != 3) badquartet = TRUE;
								} else { /* diff23 min */
									/* lowestml should be 1 */
									if (lowestml != 1) badquartet = TRUE;
								}
							} else {
								if (diff13 < diff23) { /* diff13 min*/
									/* lowestml should be 2 */
									if (lowestml != 2) badquartet = TRUE;
								} else { /* diff23 min */
									/* lowestml should be 1 */
									if (lowestml != 1) badquartet = TRUE;
								}
							}
		
							/* clear highest bit if bad quartet */
							if (badquartet) {
								quartet[a][b-a-1][c-b-1][i-c-1] =
									(quartet[a][b-a-1][c-b-1][i-c-1] & 127);
								badqs++;
								if(Info_optn) {
									fprintf(stderr, "  [bad quartet: ");
									fputid(stderr, a);
									fprintf(stderr, "-");
									fputid(stderr, b);
									fprintf(stderr, "-");
									fputid(stderr, c);
									fprintf(stderr, "-");
									fputid(stderr, i);
									fprintf(stderr, "]");
								}
							};
							
							if (firstquartet) {
								firstquartet = FALSE;
								maxdiff = diff;
								startA = a;
								startB = b;
								startC = c;
								startD = i;
							}
							
							if (!badquartet) {
								if (diff > maxdiff) {
									maxdiff = diff;
									startA = a;
									startB = b;
									startC = c;
									startD = i;	
								}
							}		
						}
	if(Info_optn) fprintf(stderr, "\n");
	fprintf(stderr, "\nINFO: %lu quartets analysed - %lu bad quartets found (= %.1f %%)\n",
		nq, badqs, (double) badqs / (double) nq * 100);
}


/* check which two *leaves* out of A, B, C are closer related to each
   other than to leaf I. These two are chooseA and chooseB 
   The relationship between leaves and taxa is defined by trueID
   If the respective quartet has not yet evaluated this will done here
   lowestml is a number 1, 2, 3 if d1, d2, d3 has the lowest ML
   checkquartet returns whether a quartet is "good" or "bad"
   (checkquartet returns TRUE if the quartet is "bad") */
int checkquartet(int A, int B, int C, int I)
{	
	int input[4][2], nochange, z, ipos;
	unsigned char qresult;
	int a, b, c, i, badquartet;
	double diff12, diff13, diff23;
	
	
	input[0][0] = trueID[A];  /* taxum number */
	input[0][1] = A;          /* leaf number */
	input[1][0] = trueID[B];
	input[1][1] = B;
	input[2][0] = trueID[C];
	input[2][1] = C;
	input[3][0] = trueID[I];
	input[3][1] = I;


	/* sort after trueIDs */
	do {
		nochange = TRUE;
		if (input[2][0] > input[3][0]) {
			z = input[3][0];
			input[3][0] = input[2][0];
			input[2][0] = z;
			z = input[3][1];
			input[3][1] = input[2][1];
			input[2][1] = z;
			nochange = FALSE;
		}
		if (input[1][0] > input[2][0]) {
			z = input[2][0];
			input[2][0] = input[1][0];
			input[1][0] = z;
			z = input[2][1];
			input[2][1] = input[1][1];
			input[1][1] = z;
			nochange = FALSE;
		}
		if (input[0][0] > input[1][0]) {
			z = input[1][0];
			input[1][0] = input[0][0];
			input[0][0] = z;
			z = input[1][1];
			input[1][1] = input[0][1];
			input[0][1] = z;
			nochange = FALSE;
		}
	} while (nochange == FALSE);

	
	/* where is leaf I ? */
	ipos = 0;
	while (input[ipos][1] != I) ipos++;


	/* look at taxum quartet */
	qresult = quartet[ input[0][0] ]
				[ input[1][0] - input[0][0] - 1 ]
				[ input[2][0] - input[1][0] - 1 ]
				[ input[3][0] - input[2][0] - 1 ];
	
	/* check whether quartet branching pattern has yet to be computed  */
	if (qresult == 'K') {
		
		a = input[0][0]; /* sorted taxum numbers */
		b = input[1][0];
		c = input[2][0];
		i = input[3][0];
		
		/* compute ml for all three topologies */
		if (Approx_optn) {
				d1 = quartet_alklhd(a,b,c,i);
				d2 = quartet_alklhd(a,c,b,i);
				d3 = quartet_alklhd(a,i,b,c);
		} else {
				d1 = quartet_lklhd(a,b,c,i);
				d2 = quartet_lklhd(a,c,b,i);
				d3 = quartet_lklhd(a,i,b,c);
		}

		/* choose topology */
		if (d1 < d2) {
			if (d2 < d3) { /* d3 max */
				quartet[a][b-a-1][c-b-1][i-c-1] = 195;
				qresult = 195;
				lowestml = 3;
			} else { /* d2 max */
				quartet[a][b-a-1][c-b-1][i-c-1] = 204;
				qresult = 204;
				lowestml = 2;
			}
		} else { 
			if (d1 < d3) { /* d3 max */
				quartet[a][b-a-1][c-b-1][i-c-1] = 195;
				qresult = 195;
				lowestml = 3;
			} else { /* d1 max */
				quartet[a][b-a-1][c-b-1][i-c-1] = 240;
				qresult = 240;
				lowestml = 1;
			}
		}

		/* check whether this quartet a good quartet */

		diff12 = (d1-d2)*(d1-d2);
		diff13 = (d1-d3)*(d1-d3);
		diff23 = (d2-d3)*(d2-d3);
		badquartet = FALSE;

		if (diff12 < diff13) {
			if (diff12 < diff23) { /* diff12 min */
				/* lowestml should be 3 */
				if (lowestml != 3) badquartet = TRUE;
			} else { /* diff23 min */
				/* lowestml should be 1 */
				if (lowestml != 1) badquartet = TRUE;
			}
		} else {
			if (diff13 < diff23) { /* diff13 min*/
				/* lowestml should be 2 */
				if (lowestml != 2) badquartet = TRUE;
			} else { /* diff23 min */
				/* lowestml should be 1 */
				if (lowestml != 1) badquartet = TRUE;
			}
		}

		/* clear highest bit if bad quartet */
		if (badquartet) {
			qresult = (qresult & 127);
			quartet[a][b-a-1][c-b-1][i-c-1] = qresult;
		};
	}

	/* is this quartet good or bad ? */
	if ((qresult & 128) == 0) { /* bad quartet */
		badquartet = TRUE;
		qresult = (qresult | 128);
	} else { /* good quartet */
		badquartet = FALSE;
	}

	/* chooseA and chooseB */
	/* chooseC is the leaf that joins I */
	                 
	switch (qresult) {

		/* 11000011 */
		case 195:	if (ipos == 0 || ipos == 3) {
						chooseA = input[1][1];
						chooseB = input[2][1];
						if (ipos == 0) chooseC = input[3][1];
						else chooseC = input[0][1];
					} else {
						chooseA = input[0][1];
						chooseB = input[3][1];
						if (ipos == 1) chooseC = input[2][1];
						else chooseC = input[1][1];
					}
					break;

		/* 11001100 */
		case 204:	if (ipos == 0 || ipos == 2) {
						chooseA = input[1][1];
						chooseB = input[3][1];
						if (ipos == 0) chooseC = input[2][1];
						else chooseC = input[0][1];
					} else {
						chooseA = input[0][1];
						chooseB = input[2][1];
						if (ipos == 1) chooseC = input[3][1];
						else chooseC = input[1][1];
					}
					break;

		/* 11110000 */
		case 240:	if (ipos == 0 || ipos == 1) {
						chooseA = input[2][1];
						chooseB = input[3][1];
						if (ipos == 0) chooseC = input[1][1];
						else chooseC = input[0][1];
					} else {
						chooseA = input[0][1];
						chooseB = input[1][1];
						if (ipos == 2) chooseC = input[3][1];
						else chooseC = input[2][1];
					}
					break;
		
		default:	fprintf(stderr, "\nERROR: Program error [checkquartet]\n");
					exit(1);
	}
	
	return badquartet;
}


/* check which two *leaves* out of A, B, C are closer related to each
   other than to leaf I. These two are chooseA and chooseB 
   The relationship between leaves and taxa is defined by trueID
   !!!!!!! The quartet will *ALWAYS* be recomputed (no storage) !!!!!!
   lowestml is a number 1, 2, 3 if d1, d2, d3 has the lowest ML
   checkquartet2 returns whether a quartet is "good" or "bad"
   (checkquartet2 returns TRUE if the quartet is "bad") */
int checkquartet2(int A, int B, int C, int I)
{	
	int input[4][2], nochange, z, ipos;
	unsigned char qresult;
	int a, b, c, i, badquartet;
	double diff12, diff13, diff23;


	input[0][0] = trueID[A];  /* taxum number */
	input[0][1] = A;          /* leaf number */
	input[1][0] = trueID[B];
	input[1][1] = B;
	input[2][0] = trueID[C];
	input[2][1] = C;
	input[3][0] = trueID[I];
	input[3][1] = I;


	/* sort after trueIDs */
	do {
		nochange = TRUE;
		if (input[2][0] > input[3][0]) {
			z = input[3][0];
			input[3][0] = input[2][0];
			input[2][0] = z;
			z = input[3][1];
			input[3][1] = input[2][1];
			input[2][1] = z;
			nochange = FALSE;
		}
		if (input[1][0] > input[2][0]) {
			z = input[2][0];
			input[2][0] = input[1][0];
			input[1][0] = z;
			z = input[2][1];
			input[2][1] = input[1][1];
			input[1][1] = z;
			nochange = FALSE;
		}
		if (input[0][0] > input[1][0]) {
			z = input[1][0];
			input[1][0] = input[0][0];
			input[0][0] = z;
			z = input[1][1];
			input[1][1] = input[0][1];
			input[0][1] = z;
			nochange = FALSE;
		}
	} while (nochange == FALSE);


	/* where is leaf I ? */
	ipos = 0;
	while (input[ipos][1] != I) ipos++;


	/* compute quartet branching pattern  */
	a = input[0][0]; /* sorted taxum numbers */
	b = input[1][0];
	c = input[2][0];
	i = input[3][0];

	/* compute ml for all three topologies */
	if (Approx_optn) {
			d1 = quartet_alklhd(a,b,c,i);
			d2 = quartet_alklhd(a,c,b,i);
			d3 = quartet_alklhd(a,i,b,c);
	} else {
			d1 = quartet_lklhd(a,b,c,i);
			d2 = quartet_lklhd(a,c,b,i);
			d3 = quartet_lklhd(a,i,b,c);
	}

	/* choose topology */
	if (d1 < d2) {
		if (d2 < d3) { /* d3 max */
			qresult = 195;
			lowestml = 3;
		} else { /* d2 max */
			qresult = 204;
			lowestml = 2;
		}
	} else { 
		if (d1 < d3) { /* d3 max */
			qresult = 195;
			lowestml = 3;
		} else { /* d1 max */
			qresult = 240;
			lowestml = 1;
		}
	}

	/* check whether this quartet a good quartet */

	diff12 = (d1-d2)*(d1-d2);
	diff13 = (d1-d3)*(d1-d3);
	diff23 = (d2-d3)*(d2-d3);
	badquartet = FALSE;

	if (diff12 < diff13) {
		if (diff12 < diff23) { /* diff12 min */
			/* lowestml should be 3 */
			if (lowestml != 3) badquartet = TRUE;
		} else { /* diff23 min */
			/* lowestml should be 1 */
			if (lowestml != 1) badquartet = TRUE;
		}
	} else {
		if (diff13 < diff23) { /* diff13 min*/
			/* lowestml should be 2 */
			if (lowestml != 2) badquartet = TRUE;
		} else { /* diff23 min */
			/* lowestml should be 1 */
			if (lowestml != 1) badquartet = TRUE;
		}
	}
	

	/* chooseA and chooseB */
	/* chooseC is the leaf that joins I */

	switch (qresult) {
		
		/* 11000011 */
		case 195:	if (ipos == 0 || ipos == 3) {
						chooseA = input[1][1];
						chooseB = input[2][1];
						if (ipos == 0) chooseC = input[3][1];
						else chooseC = input[0][1];
					} else {
						chooseA = input[0][1];
						chooseB = input[3][1];
						if (ipos == 1) chooseC = input[2][1];
						else chooseC = input[1][1];
					}
					break;
		
		/* 11001100 */
		case 204:	if (ipos == 0 || ipos == 2) {
						chooseA = input[1][1];
						chooseB = input[3][1];
						if (ipos == 0) chooseC = input[2][1];
						else chooseC = input[0][1];
					} else {
						chooseA = input[0][1];
						chooseB = input[2][1];
						if (ipos == 1) chooseC = input[3][1];
						else chooseC = input[1][1];
					}
					break;

		/* 11110000 */
		case 240:	if (ipos == 0 || ipos == 1) {
						chooseA = input[2][1];
						chooseB = input[3][1];
						if (ipos == 0) chooseC = input[1][1];
						else chooseC = input[0][1];
					} else {
						chooseA = input[0][1];
						chooseB = input[1][1];
						if (ipos == 2) chooseC = input[3][1];
						else chooseC = input[2][1];
					}
					break;
		
		default:	fprintf(stderr, "\nERROR: Program error [checkquartet2]\n");
					exit(1);
	}
	
	return badquartet;
}
